from bge import constraints
from bge import logic
from bge import render
from bge.types import KX_ConstraintWrapper as GameConstraint
from bge.types import KX_GameObject as GameObject
from mathutils import Vector
from uplogic.utils import get_direction
from uplogic.utils import set_curve_points
from uplogic.utils import xrot_to
from uplogic.utils import yrot_to
from uplogic.utils import zrot_to


CONSTRAINT_TYPES = {
    'point': 0,
    'hinge': 0,
    'angular': 0,
    'conetwist': 0,
    'generic6dof': 0
}


def create_constraint(
    obj: GameObject,
    target: GameObject,
    constraint_type: int or str = 0,
    pivot: set = (0, 0, 0),
    limit: set = (0, 0, 0),
    linked_collision: bool = True,
    local: bool = True
) -> GameConstraint:
    """Wrapper function for `bge.constraints.createConstraint()`. Creates a constraint

    :param `obj`: Object the constraint will be applied to.
    :param `target`: Target for the constraint.
    :param `constraint_type`: Type of the constraint. One of `['point', 'hinge', 'angular', 'conetwist', 'generic6dof']`.
    :param `pivot`: Point of application for the constraint.
    :param `limit`: Limit movement of the object (Like a doorstop).
    :param `linked_collision`: Enable/Disable collision between obj and target.
    :param `local`: Use obj's local space.
    """
    if not local:
        pivot[0] -= obj.worldPosition.x
        pivot[1] -= obj.worldPosition.y
        pivot[2] -= obj.worldPosition.z
    return constraints.createConstraint(
        obj.getPhysicsId(),
        target.getPhysicsId(),
        constraint_type if isinstance(constraint_type, int) else CONSTRAINT_TYPES.get(constraint_type, 0),
        pivot[0],
        pivot[1],
        pivot[2],
        limit[0],
        limit[1],
        limit[2],
        0 if linked_collision else 128
    )


def remove_constraint(constraint: GameConstraint) -> None:
    """Wrapper function for `bge.constraints.removeConstraint()`. Creates a constraint

    :param `constraint`: The constraint to remove.
    """
    constraints.removeConstraint(constraint.getConstraintId())


class ULTrackTo():
    def __init__(
        self,
        game_object: GameObject,
        target: GameObject or Vector,
        axis: int = 2,
        front: int = 1,
        speed: float = 0
    ) -> None:
        self._axis = None
        self.game_object = game_object
        self.target = target
        self.front = front
        self.speed = speed
        self.axis = axis

    @property
    def axis(self):
        return self._axis

    @axis.setter
    def axis(self, val):
        if val == 0:
            self.rotate_func = xrot_to
        elif val == 1:
            self.rotate_func = yrot_to
        elif val == 2:
            self.rotate_func = zrot_to
        else:
            self.rotate_func = None
        self._axis = val
        logic.getCurrentScene().pre_draw.append(self.update)

    def remove(self):
        logic.getCurrentScene().pre_draw.remove(self.update)

    def update(self):
        if self.rotate_func:
            self.rotate_func(self.game_object, self.target, self.front, self.speed)


class ULSpring():
    """Spring Physics Constraint.

    :param `origin`: First connection point of the spring.
    :param `target`: Second connection point of the spring.
    :param `rigid_body_origin`: Object to be influenced by the spring (optional).
    :param `rigid_body_target`: Object to be influenced by the spring (optional).
    :param `stiffness`: Amount the spring will bounce back.
    :param `max_force`: Maximum force the spring will use.
    :param `use_push`: Push the objects apart when spring is compressed.
    :param `use_breaking`: Remove the constraint when spring is pulled too much.
    :param `break_threshold`: Amount of strain the spring will endure.
    :param `curve`: Set a curve object to fit the spring.
    :param `visualize`: Enable a visual representation of the spring.
    """
    def __init__(
        self,
        origin: GameObject,
        target: GameObject,
        rigid_body_origin: GameObject = None,
        rigid_body_target: GameObject = None,
        stiffness: float = 1,
        max_force: float = -1,
        distance: float = 0,
        use_push: bool = False,
        use_breaking: bool = False,
        break_threshold: float = 1,
        curve: GameObject or None = None,
        visualize: bool = False
    ) -> None:
        self.force = 0
        self.origin = origin
        self.target = target
        self.use_push = use_push
        self.rigid_body_origin = rigid_body_origin if rigid_body_origin else origin
        self.rigid_body_target = rigid_body_target if rigid_body_target else target
        self.stiffness = stiffness
        self.max_force = max_force
        self.use_breaking = use_breaking
        self.break_threshold = break_threshold
        self.visualize = visualize
        obj_dist = origin.getDistanceTo(target)
        self.distance = distance or obj_dist
        self.curve = curve
        if not use_breaking or self.distance >= obj_dist:
            logic.getCurrentScene().pre_draw.append(self.update)

    @property
    def points(self):
        return [self.origin.worldPosition, self.target.worldPosition]
    
    def remove(self):
        logic.getCurrentScene().pre_draw.remove(self.update)

    def update(self):
        o = self.origin
        t = self.target
        force = (o.getDistanceTo(t) - self.distance) * self.stiffness
        if self.max_force >= 0 and force > self.max_force:
            force = self.max_force
        if not self.use_push:
            force = force if force >= 0 else 0
        if self.use_breaking and force > self.break_threshold:
            logic.getCurrentScene().pre_draw.remove(self.update)
            return
        if self.visualize:
            render.drawLine(
                o.worldPosition,
                t.worldPosition,
                # [1, 1-abs(power), 1-abs(power)]
                [abs(force), 0, 1-abs(force)]
            )
        self.force = force
        if self.curve:
            set_curve_points(self.curve, self.points)
        if self.rigid_body_origin.blenderObject.data:
            self.rigid_body_origin.applyImpulse(o.worldPosition, get_direction(o, t) * force)
        if self.rigid_body_target.blenderObject.data:
            self.rigid_body_target.applyImpulse(o.worldPosition, get_direction(t, o) * force)


from bge import logic
from uplogic.nodes import ULParameterNode
from uplogic.nodes import ULOutSocket


class ULGetTimeScale(ULParameterNode):
    def __init__(self):
        ULParameterNode.__init__(self)
        self.OUT = ULOutSocket(self, self.get_done)

    def get_done(self):
        return logic.getTimeScale()

    def evaluate(self):
        self._set_ready()



'''TODO: Documentation
'''

from .curves import set_curve_points  # noqa
from .curves import create_curve  # noqa
from .raycasting import raycast
from .raycasting import raycast_projectile
from .raycasting import raycast_camera

from bge import logic, render
from bge.types import KX_GameObject as GameObject
from mathutils import Vector
from mathutils import Matrix
import bpy
import json
import math
import operator


###############################################################################
# CONSTANTS
###############################################################################

class _Status(object):
    def __init__(self, name):
        self._name = name

    def __repr__(self):
        return self._name


NO_VALUE = _Status("NO_VALUE")
STATUS_WAITING = _Status("WAITING")
STATUS_READY = _Status("READY")
STATUS_INVALID = _Status("INVALID")


# uplogic game properties
VEHICLE = '.ulvehicleconst'
SHIP = '.ulshipconst'
FLOATSAM = '.ulfloatsamconst'

WATER = '.ulwater'


LOGIC_OPERATORS = [
    operator.eq,
    operator.ne,
    operator.gt,
    operator.lt,
    operator.ge,
    operator.le
]


LO_AXIS_TO_STRING_CODE = {
    0: "X", 1: "Y", 2: "Z",
    3: "-X", 4: "-Y", 5: "-Z",
}


LO_AXIS_TO_VECTOR = {
    0: Vector((1, 0, 0)), 1: Vector((0, 1, 0)),
    2: Vector((0, 0, 1)), 3: Vector((-1, 0, 0)),
    4: Vector((0, -1, 0)), 5: Vector((0, 0, -1)),
}


###############################################################################
# LOGIC NODES
###############################################################################


def _name_query(named_items, query):
    assert len(query) > 0
    postfix = (query[0] == "*")
    prefix = (query[-1] == "*")
    infix = (prefix and postfix)
    if infix:
        token = query[1:-1]
        for item in named_items:
            if token in item.name:
                return item
    if prefix:
        token = query[:-1]
        for item in named_items:
            if item.name.startswith(token):
                return item
    if postfix:
        token = query[1:]
        for item in named_items:
            if item.name.endswith(token):
                return item
    for item in named_items:
        if item.name == query:
            return item
    return None


def check_game_object(query, scene=None):
    '''TODO: Documentation
    '''
    if not scene:
        scene = logic.getCurrentScene()
    else:
        scene = scene
    if (query is None) or (query == ""):
        return
    if not is_invalid(scene):
        # find from scene
        return _name_query(scene.objects, query)


def compute_distance(parama, paramb) -> float:
    '''TODO: Documentation
    '''
    if is_invalid(parama):
        return None
    if is_invalid(paramb):
        return None
    if hasattr(parama, "getDistanceTo"):
        return parama.getDistanceTo(paramb)
    if hasattr(paramb, "getDistanceTo"):
        return paramb.getDistanceTo(parama)
    va = Vector(parama)
    vb = Vector(paramb)
    return (va - vb).length


def debug(message: str):
    if not hasattr(bpy.types.Scene, 'logic_node_settings'):
        return
    if not bpy.context or not bpy.context.scene:
        return
    if not bpy.context.scene.logic_node_settings.use_node_debug:
        return
    else:
        print('[UPLOGIC] ' + message)


def is_invalid(*a) -> bool:
    for ref in a:
        if ref is None or ref is STATUS_WAITING or ref == '':
            return True
        if not hasattr(ref, "invalid"):
            continue
        elif ref.invalid:
            return True
    return False


def is_waiting(*args) -> bool:
    if STATUS_WAITING in args:
        return True
    return False


def make_valid_name(name):
    valid_characters = (
        "abcdefghijklmnopqrstuvwxyz1234567890ABCDEFGHIJKLMNOPQRSTUVWXYZ"
    )
    clsname = name.replace(' ', '_')
    stripped_name = "".join(
        [c for c in clsname if c in valid_characters]
    )
    return stripped_name


def not_met(*conditions) -> bool:
    for c in conditions:
        if (
            c is STATUS_WAITING or
            c is None or
            c is False
        ):
            return True
    return False


def load_user_module(module_name):
    import sys
    exec(f"import {module_name}")
    return sys.modules[module_name]


def unload_nodes(a, b):
    if not hasattr(bpy.types.Scene, 'nl_globals_initialized'):
        return
    delattr(bpy.types.Scene, 'nl_globals_initialized')


def move_to(
    moving_object,
    destination_point,
    speed,
    time_per_frame,
    dynamic,
    distance
):
    if dynamic:
        direction = (
            destination_point -
            moving_object.worldPosition)
        dst = direction.length
        if(dst <= distance):
            return True
        direction.z = 0
        direction.normalize()
        velocity = direction * speed
        velocity.z = moving_object.worldLinearVelocity.z
        moving_object.worldLinearVelocity = velocity
        return False
    else:
        direction = (
            destination_point -
            moving_object.worldPosition
            )
        dst = direction.length
        if(dst <= distance):
            return True
        direction.normalize()
        displacement = speed * time_per_frame
        motion = direction * displacement
        moving_object.worldPosition += motion
        return False


def project_vector3(v, xi, yi):
    return Vector((v[xi], v[yi]))


def xrot_to(
    rotating_object,
    target_pos,
    front_axis_code=0,
    speed=0,
    time_per_frame=1/60
):
    front_vector = LO_AXIS_TO_VECTOR[front_axis_code]
    vec = rotating_object.getVectTo(target_pos)[1]
    if speed == 0:
        if front_axis_code >= 3:
            vec.negate()
            front_axis_code = front_axis_code - 3
        if vec.x == vec.y == vec.z == 0:
            return
        rotating_object.alignAxisToVect(vec, front_axis_code, 1.0)
        rotating_object.alignAxisToVect(LO_AXIS_TO_VECTOR[0], 0, 1.0)
        return True
    else:
        vec = project_vector3(vec, 1, 2)
        vec.normalize()
        front_vector = rotating_object.getAxisVect(front_vector)
        front_vector = project_vector3(front_vector, 1, 2)
        signed_angle = vec.angle_signed(front_vector, None)
        if signed_angle is None:
            return
        abs_angle = abs(signed_angle)
        if abs_angle < 0.01:
            return True
        angle_sign = (signed_angle > 0) - (signed_angle < 0)
        drot = angle_sign * abs_angle * speed * time_per_frame
        eulers = rotating_object.localOrientation.to_euler()
        eulers[0] += drot
        rotating_object.localOrientation = eulers
        return False


def yrot_to(
    rotating_object,
    target_pos,
    front_axis_code=0,
    speed=0,
    time_per_frame=1/60
):
    front_vector = LO_AXIS_TO_VECTOR[front_axis_code]
    vec = rotating_object.getVectTo(target_pos)[1]
    if speed == 0:
        if front_axis_code >= 3:
            vec.negate()
            front_axis_code = front_axis_code - 3
        if vec.x == vec.y == vec.z == 0:
            return
        rotating_object.alignAxisToVect(vec, front_axis_code, 1.0)
        rotating_object.alignAxisToVect(LO_AXIS_TO_VECTOR[1], 1, 1.0)
        return True
    else:
        vec = project_vector3(vec, 2, 0)
        vec.normalize()
        front_vector = rotating_object.getAxisVect(front_vector)
        front_vector = project_vector3(front_vector, 2, 0)
        signed_angle = vec.angle_signed(front_vector, None)
        if signed_angle is None:
            return
        abs_angle = abs(signed_angle)
        if abs_angle < 0.01:
            return True
        angle_sign = (signed_angle > 0) - (signed_angle < 0)
        drot = angle_sign * abs_angle * speed * time_per_frame
        eulers = rotating_object.localOrientation.to_euler()
        eulers[1] += drot
        rotating_object.localOrientation = eulers
        return False


def zrot_to(
    rotating_object,
    target_pos,
    front_axis_code=0,
    speed=0,
    time_per_frame=1/60
):
    front_vector = LO_AXIS_TO_VECTOR[front_axis_code]
    vec = rotating_object.getVectTo(target_pos)[1]
    if speed == 0:
        if front_axis_code >= 3:
            vec.negate()
            front_axis_code = front_axis_code - 3
        if vec.x == vec.y == vec.z == 0:
            return
        rotating_object.alignAxisToVect(vec, front_axis_code, 1.0)
        rotating_object.alignAxisToVect(LO_AXIS_TO_VECTOR[2], 2, 1.0)
        return True
    else:
        # project in 2d, compute angle diff, set euler rot 2
        vec = project_vector3(vec, 0, 1)
        vec.normalize()
        front_vector = rotating_object.getAxisVect(front_vector)
        front_vector = project_vector3(front_vector, 0, 1)
        signed_angle = vec.angle_signed(front_vector, None)
        if signed_angle is None:
            return True
        abs_angle = abs(signed_angle)
        if abs_angle < 0.01:
            return True
        angle_sign = (signed_angle > 0) - (signed_angle < 0)
        drot = angle_sign * abs_angle * speed * time_per_frame
        eulers = rotating_object.localOrientation.to_euler()
        eulers[2] += drot
        rotating_object.localOrientation = eulers
        return False


def rot_to(
    rot_axis_index,
    rotating_object,
    target_pos,
    front_axis_code,
    speed,
    time_per_frame
):
    if rot_axis_index == 0:
        return xrot_to(
            rotating_object,
            target_pos,
            front_axis_code,
            speed,
            time_per_frame
        )
    elif rot_axis_index == 1:
        return yrot_to(
            rotating_object,
            target_pos,
            front_axis_code,
            speed,
            time_per_frame
        )
    elif rot_axis_index == 2:
        return zrot_to(
            rotating_object,
            target_pos,
            front_axis_code,
            speed,
            time_per_frame
        )


###############################################################################
# DATA
###############################################################################

def load_json_as_dict(filepath):
    if not filepath.endswith('.json'):
        path = f'{filepath}.json'
    if path:
        f = open(path, 'r')
        data = json.load(f)
        f.close()
        return data


###############################################################################
# SCENE
###############################################################################

def make_unique_light(old_lamp_ge: GameObject) -> GameObject:
    '''TODO: Documentation
    '''
    scene = logic.getCurrentScene()
    name = old_lamp_ge.name
    old_lamp = old_lamp_ge.blenderObject
    old_lamp.name = 'OLD_LAMP'

    settings = [
        'color',
        'energy',
        'diffuse_factor',
        'specular_factor',
        'volume_factor',
        'shadow_soft_size',
        'use_custom_distance',
        'cutoff_distance',
        'spot_size',
        'spot_blend',
        'show_cone',
        'angle',
        'shape',
        'size',
        'size_y',
        'use_shadow',
        'shadow_buffer_clip_start',
        'shadow_buffer_bias',
        'use_soft_shadows',
        'use_contact_shadow',
        'contact_shadow_distance',
        'contact_shadow_bias',
        'contact_shadow_thickness',
        'shadow_cascade_count',
        'shadow_cascade_fade',
        'shadow_cascade_max_distance',
        'shadow_cascade_exponent'
    ]

    types = {
        'POINT': 'Point',
        'AREA': 'Area',
        'SPOT': 'Spot',
        'SUN': 'Sun'
    }

    light_type = old_lamp.data.type
    bpy.ops.object.light_add(
        type=light_type,
        location=old_lamp_ge.worldPosition,
        rotation=old_lamp_ge.worldOrientation.to_euler()
    )
    index = 1
    light = None
    while light is None:
        if types[light_type] in bpy.data.objects[-index].name:
            light = bpy.data.objects[-index]
        index += 1
    for attr in settings:
        try:
            setattr(light.data, attr, getattr(old_lamp.data, attr))
        except Exception:
            pass
    light.name = name
    new_lamp_ge = scene.convertBlenderObject(light)
    for p in old_lamp_ge.getPropertyNames():
        new_lamp_ge[p] = old_lamp_ge[p]
    old_lamp_ge.endObject()
    real_name = light.name
    light.name = real_name
    if old_lamp_ge.parent:
        new_lamp_ge.setParent(old_lamp_ge.parent)
    return new_lamp_ge


def get_closest_instance(game_obj: GameObject, name: str):
    '''TODO: Documentation
    '''
    objs = []
    distances = {}
    for obj in logic.getCurrentScene().objects:
        if obj.name == name:
            objs.append(obj)
    for obj in objs:
        distances[game_obj.getDistanceTo(obj)] = obj
    return distances[min(distances.keys())]


def is_water(game_object: GameObject):
    return WATER in game_object


def get_child_by_name(obj, child, recursive=True, partial=False):
    children = obj.childrenRecursive if recursive else obj.children
    if partial:
        for c in children:
            if child in c.name:
                return c
    else:
        return children.get(child)


###############################################################################
# MATH
###############################################################################


def clamp(value: float, min: float = 0, max: float = 1) -> float:
    """Clamp a value in between two other values.

    :param value: input value
    :param min: minimum value
    :param max: maximum value

    :returns: clamped value as float
    """

    if value < min:
        return min
    if value > max:
        return max
    return value


def vec_clamp(vec: Vector, min: float = 0, max: float = 1) -> Vector:
    """Clamp length of a vector.

    :param value: `Vector`
    :param min: minimum length
    :param max: maximum length

    :returns: clamped vector as float
    """
    vec = vec.copy()

    if vec.length < min:
        vec.normalize()
        return vec * min
    if vec.length > max:
        vec.normalize()
        return vec * max
    return vec


def interpolate(a: float, b: float, fac: float, threshold: float = 0.001) -> float:
    """Interpolate between 2 values using a factor.

    :param a: starting value
    :param b: target value
    :param fac: interpolation factor

    :returns: calculated value as float
    """
    if -threshold < a-b < threshold:
        return b
    return (fac * b) + ((1-fac) * a)


def lerp(a: float, b: float, fac: float) -> float:
    """Interpolate between 2 values using a factor.

    :param a: starting value
    :param b: target value
    :param fac: interpolation factor

    :returns: calculated value as float
    """
    if -.001 < a-b < .001:
        return b
    return (fac * b) + ((1-fac) * a)


def vec_abs(vec):
    """Make every value of this vector positive.\n
    Only supports less than 4 Dimensions.

    :param `a`: `Vector`

    :returns: positive vector
    """
    vec = vec.copy()
    vec.x = abs(vec.x)
    vec.y = abs(vec.y)
    vec.z = abs(vec.z)
    return vec


def get_angle(a: Vector, b: Vector, up=Vector((0, 0, 1))) -> float:
    """Get the angle between the direction from a to b and up.

    :param `a`: `Vector` a
    :param `b`: `Vector` b
    :param `up`: compare direction

    :returns: calculated value as float
    """
    direction = get_direction(a, b)
    rad: float = direction.angle(up)
    deg: float = rad * 180/math.pi
    return deg


def get_raw_angle(a: Vector, b: Vector) -> float:
    """Get the angle between the direction from a to b and up.

    :param `a`: `Vector` a
    :param `b`: `Vector` b
    :param `up`: compare direction

    :returns: calculated value as float
    """
    rad: float = a.angle(b)
    deg: float = rad * 180/math.pi
    return deg


def get_direction(a, b, local=False) -> Vector:
    """Get the direction from one vector to another.

    :param `a`: `Vector` a
    :param `b`: `Vector` b
    :param `local`: use local space (position only)

    :returns: direction as `Vector`
    """
    start = a.worldPosition.copy() if hasattr(a, "worldPosition") else a
    if hasattr(b, "worldPosition"):
        b = b.worldPosition.copy()
    if local:
        b = start + b
    d = b - start
    d.normalize()
    return d



from uplogic.nodes import ULParameterNode
from uplogic.nodes import ULOutSocket
from uplogic.utils import STATUS_WAITING
from uplogic.utils import is_waiting
from uplogic.utils import is_invalid


class ULVectorMath(ULParameterNode):
    def __init__(self):
        ULParameterNode.__init__(self)
        self.op = None
        self.vector = None
        self.vector_2 = None
        self.factor = None
        self.OUT = ULOutSocket(self, self.get_done)

    def get_done(self):
        op = self.get_input(self.op)
        vector = self.get_input(self.vector)
        vector_2 = self.get_input(self.vector_2)
        factor = self.get_input(self.factor)
        if is_waiting(
            op,
            factor
        ):
            return STATUS_WAITING
        if is_invalid(
            vector,
            vector_2
        ):
            return STATUS_WAITING
        return self.calc_output_vector(op, vector, vector_2, factor)

    def evaluate(self):
        self._set_ready()

    def calc_output_vector(self, op, vec, vec2, fac):
        matvec = vec.copy()
        if op == 'normalize':
            matvec.normalize()
        elif op == 'lerp':
            return matvec.lerp(vec2, fac)
        elif op == 'negate':
            matvec.negate()
        elif op == 'dot':
            return matvec.dot(vec2)
        elif op == 'cross':
            return matvec.cross(vec2)
        elif op == 'project':
            return matvec.project(vec2)
        return matvec



from mathutils import Vector
from uplogic.nodes import ULOutSocket
from uplogic.nodes import ULParameterNode
from uplogic.utils import is_invalid


class ULGetCurvePoints(ULParameterNode):
    def __init__(self):
        ULParameterNode.__init__(self)
        self.curve = None
        self.OUT = ULOutSocket(self, self.get_points)

    def get_points(self):
        socket = self.get_output('points')
        if socket is None:
            obj = self.get_input(self.curve)
            if is_invalid(obj):
                return
            offset = obj.worldPosition
            o = obj.blenderObject.data.splines[0]
            if o.type == 'BEZIER':
                return self.set_output(
                    'points',
                    [Vector(p.co) + offset for p in o.bezier_points]
                )
            return self.set_output(
                'points',
                [Vector(p.co[:-1]) + offset for p in o.points]
            )
        return socket

    def evaluate(self):
        self._set_ready()



from uplogic.audio import ULSound3D
from uplogic.nodes import ULActionNode
from uplogic.nodes import ULOutSocket
from uplogic.utils import is_invalid
from uplogic.utils import not_met


class ULStartSound3D(ULActionNode):
    def __init__(self):
        ULActionNode.__init__(self)
        self.condition = None
        self.sound = None
        self.occlusion = None
        self.reverb = None
        self.transition = None
        self.cutoff = None
        self.speaker = None
        self.device = None
        self.loop_count = None
        self.pitch = None
        self.volume = None
        self.attenuation = None
        self.distance_ref = None
        self.cone_angle = None
        self.cone_outer_volume = None
        self.done = None
        self.on_finish = False
        self._clear_sound = 1
        self._sustained = 1
        self._handle = None
        self.DONE = ULOutSocket(self, self.get_done)
        self.ON_FINISH = ULOutSocket(self, self.get_on_finish)
        self.HANDLE = ULOutSocket(self, self.get_handle)

    def get_handle(self):
        return self._handle

    def get_on_finish(self):
        if not self._handle:
            return False
        if self._handle.finished:
            return True
        return False

    def get_done(self):
        return self.done
    
    def reset(self):
        ULActionNode.reset(self)
        if self._handle and self._handle.finished:
            self._handle = None

    def evaluate(self):
        self.done = False
        self.on_finish = False
        volume = self.get_input(self.volume)
        pitch = self.get_input(self.pitch)
        if self._handle:
            self._handle.volume = volume
            self._handle.pitch = pitch
        condition = self.get_input(self.condition)
        if not_met(condition):
            self._set_ready()
            return
        speaker = self.get_input(self.speaker)
        transition = self.get_input(self.transition)
        reverb = self.get_input(self.reverb)
        occlusion = self.get_input(self.occlusion)
        cone_outer_volume = self.get_input(self.cone_outer_volume)
        attenuation = self.get_input(self.attenuation)
        cutoff = self.get_input(self.cutoff)
        file = self.get_input(self.sound)
        loop_count = self.get_input(self.loop_count)
        distance_ref = self.get_input(self.distance_ref)
        cone_angle = self.get_input(self.cone_angle)
        self._set_ready()

        if is_invalid(file):
            return
        self._handle = ULSound3D(
            speaker,
            file,
            occlusion,
            transition,
            cutoff,
            loop_count,
            pitch,
            volume,
            reverb,
            attenuation,
            distance_ref,
            [cone_angle.x, cone_angle.y],
            cone_outer_volume,
            'default'
        )
        self.done = True



from uplogic.nodes import ULActionNode
from uplogic.nodes import ULOutSocket
from uplogic.utils import is_waiting, make_valid_name
from uplogic.utils import is_invalid
from uplogic.utils import not_met


class ULStopSubNetwork(ULActionNode):
    def __init__(self):
        ULActionNode.__init__(self)
        self.condition = None
        self.game_object = None
        self.logic_network_name = None
        self.done = None
        self.OUT = ULOutSocket(self, self.get_done)

    def get_done(self):
        return self.done

    def evaluate(self):
        self.done = False
        condition = self.get_input(self.condition)
        if not_met(condition):
            return
        game_object = self.get_input(self.game_object)
        logic_network_name = self.get_input(self.logic_network_name)
        if is_waiting(game_object, logic_network_name):
            return
        self._set_ready()
        if is_invalid(game_object):
            return
        tree_name = make_valid_name(logic_network_name)
        network = game_object.get(f'IGNLTree_{tree_name}')
        if network is None:
            return
        network.stop()
        self.done = True



from .gamepad import LS  # noqa
from .gamepad import RS  # noqa
from .gamepad import SONY  # noqa
from .gamepad import XBOX  # noqa
from .gamepad import gamepad_down  # noqa
from .gamepad import gamepad_stick  # noqa
from .gamepad import gamepad_tap  # noqa
from .gamepad import gamepad_up  # noqa
from .keyboard import KEYBOARD_EVENTS  # noqa
from .keyboard import key_down  # noqa
from .keyboard import key_pulse  # noqa
from .keyboard import key_tap  # noqa
from .mouse import LMB  # noqa
from .mouse import MMB  # noqa
from .mouse import MOUSE_EVENTS  # noqa
from .mouse import RMB  # noqa
from .mouse import ULMouseData  # noqa
from .mouse import get_mouse_position  # noqa
from .mouse import mouse_down  # noqa
from .mouse import mouse_moved  # noqa
from .mouse import mouse_tap  # noqa
from .mouse import mouse_wheel  # noqa
from .mouse import set_mouse_position  # noqa
from .mouse import ULMouseLook  # noqa



from uplogic.nodes import ULActionNode
from uplogic.nodes import ULOutSocket
from uplogic.utils import is_invalid
from uplogic.utils import is_waiting
from uplogic.utils import not_met
from uplogic.utils import debug
import bpy


class ULSetMaterial(ULActionNode):
    def __init__(self):
        ULActionNode.__init__(self)
        self.condition = None
        self.game_object = None
        self.slot = None
        self.mat_name = None
        self.done = False
        self.OUT = ULOutSocket(self, self._get_done)

    def _get_done(self):
        return self.done

    def evaluate(self):
        self.done = False
        condition = self.get_input(self.condition)
        game_object = self.get_input(self.game_object)
        slot = self.get_input(self.slot) - 1
        mat_name = self.get_input(self.mat_name)
        if not_met(condition):
            return
        if is_invalid(game_object):
            return
        if is_waiting(mat_name, slot):
            return
        self._set_ready()
        bl_obj = game_object.blenderObject
        if slot > len(bl_obj.material_slots) - 1:
            debug('Set Material: Slot does not exist!')
            return
        bl_obj.material_slots[slot].material = bpy.data.materials[mat_name]
        self.done = True



from uplogic.data import GlobalDB
from uplogic.nodes import ULActionNode
from uplogic.nodes import ULOutSocket
from uplogic.utils import is_waiting
from uplogic.utils import not_met


class ULSetGlobalValue(ULActionNode):
    def __init__(self):
        ULActionNode.__init__(self)
        self.condition = None
        self.data_id = None
        self.key = None
        self.value = None
        self.persistent = None
        self.done = None
        self.OUT = ULOutSocket(self, self.get_done)

    def get_done(self):
        return self.done

    def evaluate(self):
        self.done = False
        condition = self.get_input(self.condition)
        if not_met(condition):
            self._set_ready()
            return
        data_id = self.get_input(self.data_id)
        persistent = self.get_input(self.persistent)
        key = self.get_input(self.key)
        value = self.get_input(self.value)
        if is_waiting(data_id, persistent, key, value):
            return
        self._set_ready()
        if self.condition is None or condition:
            if data_id is None:
                return
            if persistent is None:
                return
            if key is None:
                return
            db = GlobalDB.retrieve(data_id)
            db.put(key, value, persistent)
            if self.condition is None:
                self.deactivate()
        self.done = True



from uplogic.nodes import ULOutSocket, ULParameterNode
from uplogic.utils import STATUS_WAITING, is_invalid


class ULGetProperty(ULParameterNode):
    def __init__(self):
        ULParameterNode.__init__(self)
        self.game_object = None
        self.property_name = None
        self.OUT = ULOutSocket(self, self.get_property)

    def get_property(self):
        game_object = self.get_input(self.game_object)
        property_name = self.get_input(self.property_name)
        if is_invalid(game_object, property_name):
            return STATUS_WAITING
        if property_name in game_object:
            return game_object[property_name]
        game_object[property_name] = None

    def evaluate(self):
        self._set_ready()



from uplogic.nodes import ULOutSocket
from uplogic.nodes import ULParameterNode


class ULInitEmptyDict(ULParameterNode):
    def __init__(self):
        ULParameterNode.__init__(self)
        self.condition = None
        self.dict = None
        self.done = None
        self.DICT = ULOutSocket(self, self.get_dict)

    def get_dict(self):
        return {}

    def evaluate(self):
        self._set_ready()



from mathutils import Euler
from mathutils import Vector
from uplogic.nodes import ULActionNode
from uplogic.nodes import ULOutSocket
from uplogic.utils import debug
from uplogic.utils import is_invalid
from uplogic.utils import not_met


class ULSetBonePosition(ULActionNode):
    def __init__(self):
        ULActionNode.__init__(self)
        self.condition = None
        self.armature = None
        self.bone_name = None
        self.set_translation = None
        self._eulers = Euler((0, 0, 0), "XYZ")
        self._vector = Vector((0, 0, 0))
        self.done = None
        self.OUT = ULOutSocket(self, self.get_done)

    def get_done(self):
        return self.done

    def evaluate(self):
        self.done = False
        condition = self.get_input(self.condition)
        if not_met(condition):
            return
        armature = self.get_input(self.armature)
        bone_name = self.get_input(self.bone_name)
        set_translation = self.get_input(self.set_translation)
        self._set_ready()
        if is_invalid(armature, bone_name, set_translation):
            return
        if not bone_name:
            return
        # TODO cache the bone index
        bone_channel = armature.channels[bone_name]
        if set_translation is not None:
            bone_channel.location = set_translation
        else:
            debug('Set Bone Node: Position is None!')
        armature.update()
        self.done = True



from bge import logic
from uplogic.nodes import ULActionNode
from uplogic.nodes import ULOutSocket
from uplogic.utils import is_waiting
from uplogic.utils import not_met


class ULAddObject(ULActionNode):
    def __init__(self):
        ULActionNode.__init__(self)
        self.condition = None
        self.name = None
        self.reference = None
        self.life = None
        self.full_copy = False
        self.done = False
        self.obj = False
        self.OBJ = ULOutSocket(self, self._get_obj)
        self.OUT = ULOutSocket(self, self._get_done)

    def _get_done(self):
        return self.done

    def _get_obj(self):
        return self.obj

    def evaluate(self):
        self.done = False
        condition = self.get_input(self.condition)
        if not_met(condition):
            return
        life = self.get_input(self.life)
        name = self.get_input(self.name)
        full_copy = self.get_input(self.full_copy)
        self._set_ready()
        reference = self.get_input(self.reference)
        scene = logic.getCurrentScene()
        if is_waiting(life, name, reference):
            return
        self.obj = scene.addObject(name, reference, life, full_copy)
        self.done = True



from bge import events
from uplogic.nodes import ULConditionNode


class ULMouseMoved(ULConditionNode):
    def __init__(self):
        ULConditionNode.__init__(self)
        self.pulse = False

    def evaluate(self):
        self._set_ready()
        mstat = self.network.mouse_events
        if self.pulse:
            self._set_value(
                mstat[events.MOUSEX].active or
                mstat[events.MOUSEX].activated or
                mstat[events.MOUSEY].active or
                mstat[events.MOUSEY].activated
            )
        else:
            self._set_value(
                mstat[events.MOUSEX].activated or
                mstat[events.MOUSEY].activated
            )



from bge import logic
from uplogic.nodes import ULActionNode
from uplogic.nodes import ULOutSocket
from uplogic.utils import is_invalid
from uplogic.utils import not_met


class ULCursorBehavior(ULActionNode):
    def __init__(self):
        ULActionNode.__init__(self)
        self.condition = None
        self.cursor_object = None
        self.world_z = None
        self.done = None
        self.OUT = ULOutSocket(self, self.get_done)

    def get_done(self):
        return self.done

    def evaluate(self):
        self.done = False
        self._set_ready()
        cursor_object = self.get_input(self.cursor_object)
        if is_invalid(cursor_object):
            return
        condition = self.get_input(self.condition)
        if not_met(condition):
            if cursor_object.visible:
                cursor_object.setVisible(False, True)
            return
        camera = logic.getCurrentScene().active_camera
        world_z = self.get_input(self.world_z)
        if not cursor_object.visible:
            cursor_object.setVisible(True, True)
        else:
            x = self.network._last_mouse_position[0]
            y = self.network._last_mouse_position[1]
            direction = camera.getScreenVect(x, y)
            origin = camera.worldPosition
            aim = direction * -world_z
            point = origin + aim
            cursor_object.worldOrientation = camera.worldOrientation
            cursor_object.worldPosition = point
        self.done = True



from math import pi
from bge import logic
from bge import events
from bge import render
from bge.types import KX_GameObject as GameObject
from mathutils import Vector
from uplogic.utils import interpolate
from uplogic.events import schedule_callback


MOUSE_EVENTS = logic.mouse.inputs
'''Reference to `bge.logic.mouse.inputs`
'''

LMB = events.LEFTMOUSE
RMB = events.RIGHTMOUSE
MMB = events.MIDDLEMOUSE


class ULMouseData():
    def __init__(self) -> None:
        self.position = get_mouse_position()
        self.movement = (0, 0)
        self.wheel = mouse_wheel()
        logic.getCurrentScene().pre_draw.append(self.update)

    def update(self):
        old_pos = self.position
        new_pos = get_mouse_position()
        self.movement = (
            new_pos[0] - old_pos[0],
            new_pos[1] - old_pos[1]
        )
        self.position = new_pos
        self.wheel = mouse_wheel()

    def destroy(self):
        logic.getCurrentScene().pre_draw.remove(self.update)


def set_mouse_position(x: int, y: int, absolute: bool = False):
    if absolute:
        render.setMousePosition(x, y)
        return
    render.setMousePosition(
        int(x * render.getWindowWidth()),
        int(y * render.getWindowHeight())
    )


def get_mouse_position(absolute: bool = False):
    pos = logic.mouse.position
    if absolute:
        return (
            int(pos[0] * render.getWindowWidth()),
            int(pos[1] * render.getWindowHeight())
        )
    return pos


def mouse_moved(tap: bool = False) -> bool:
    '''Detect mouse movement.

    :param tap: Only use the first consecutive `True` output

    :returns: boolean
    '''
    if tap:
        return (
            MOUSE_EVENTS[events.MOUSEX].activated or
            MOUSE_EVENTS[events.MOUSEY].activated
        )
    else:
        return (
            MOUSE_EVENTS[events.MOUSEX].active or
            MOUSE_EVENTS[events.MOUSEY].active or
            MOUSE_EVENTS[events.MOUSEX].activated or
            MOUSE_EVENTS[events.MOUSEY].activated
        )


def mouse_tap(button=events.LEFTMOUSE) -> bool:
    '''Detect mouse button tap.

    :param button: can be either `LMB`, `RMB` or `MMB` from `uplogic.input`

    :returns: boolean
    '''
    return (
        MOUSE_EVENTS[button].activated or
        MOUSE_EVENTS[button].activated
    )


def mouse_down(button=events.LEFTMOUSE) -> bool:
    '''Detect mouse button held down.

    :param button: can be either `LMB`, `RMB` or `MMB` from `uplogic.input`

    :returns: boolean
    '''
    return (
        MOUSE_EVENTS[button].active or
        MOUSE_EVENTS[button].activated or
        MOUSE_EVENTS[button].active or
        MOUSE_EVENTS[button].activated
    )


def mouse_up(button=events.LEFTMOUSE) -> bool:
    '''Detect mouse button released.

    :param button: can be either `LMB`, `RMB` or `MMB` from `uplogic.input`

    :returns: boolean
    '''
    return (
        MOUSE_EVENTS[button].released
    )


def mouse_wheel(tap: bool = False) -> int:
    '''Detect mouse wheel activity.

    :param tap: Only use the first consecutive `True` output

    :returns: -1 if wheel down, 0 if idle, 1 if wheel up
    '''
    if tap:
        return (
            MOUSE_EVENTS[events.WHEELUPMOUSE].activated -
            MOUSE_EVENTS[events.WHEELDOWNMOUSE].activated
        )
    else:
        return (
            (
                MOUSE_EVENTS[events.WHEELUPMOUSE].activated or
                MOUSE_EVENTS[events.WHEELUPMOUSE].active
            ) - (
                MOUSE_EVENTS[events.WHEELDOWNMOUSE].activated or
                MOUSE_EVENTS[events.WHEELDOWNMOUSE].active
            )
        )


class ULMouseLook():

    def __init__(
        self,
        obj: GameObject,
        head=None,
        sensitivity=1.0,
        use_cap_x=False,
        cap_x=[0, 0],
        use_cap_y=False,
        cap_y=[-89, 89],
        invert=[False, True],
        smoothing=0.0,
        local=False,
        front=1,
        active=True
    ) -> None:
        self.obj = obj
        self.head = head if head else obj
        self._defaults = [obj.localOrientation.copy(), head.localOrientation.copy()]
        self.sensitivity = sensitivity
        self.use_cap_x = use_cap_x
        self.cap_x = cap_x
        self.use_cap_y = use_cap_y
        self.cap_y = cap_y
        self.invert = invert
        self.smoothing = smoothing
        self.initialized = False
        self.front = front
        self._x = 0
        self._y = 0
        self._active = False
        self.local = local
        self.reset_factor = 0
        self.get_data()
        self.mouse.position = self.screen_center
        self.active = active

    @property
    def active(self):
        return self._active

    @active.setter
    def active(self, val):
        scene = logic.getCurrentScene()
        if val and self.update not in scene.pre_draw:
            scene.pre_draw.append(self.update)
        elif not val and self.update in scene.pre_draw:
            self.initialized = False
            scene.pre_draw.remove(self.update)
        self._active = val

    @property
    def rotation(self):
        return self.obj.worldOrientation, self.head.worldOrientation

    @rotation.setter
    def rotation(self, val):
        self.obj.worldOrientation = val[0]
        self.head.worldOrientation = val[1]
    
    def stop(self):
        self.active = False
        self.initialized = False

    def reset(self, factor=1):
        if factor < 1:
            self.active = False
            if self.reset_factor < 1:
                self.obj.localOrientation = self.obj.localOrientation.lerp(self._defaults[0], factor)
                self.head.localOrientation = self.head.localOrientation.lerp(self._defaults[1], factor)
                self.reset_factor = interpolate(self.reset_factor, 1, factor)
                schedule_callback(self.reset, arg=factor)
            else:
                self.reset_factor = 0
                self.reset()
        else:
            self.obj.localOrientation = self._defaults[0]
            self.head.localOrientation = self._defaults[1]

    def get_data(self):
        self.x = render.getWindowWidth()//2
        self.y = render.getWindowHeight()//2
        self.screen_center = (
            self.x / render.getWindowWidth(),
            self.y / render.getWindowHeight()
        )
        self.center = Vector(self.screen_center)
        self.mouse = logic.mouse

    def update(self):
        self.get_data()
        if not self.initialized:
            self.mouse.position = self.screen_center
            self.initialized = True
            return
        game_object_x = self.obj
        game_object_y = self.head
        sensitivity = self.sensitivity * 10
        use_cap_x = self.use_cap_x
        use_cap_y = self.use_cap_y
        cap_x = self.cap_x
        lowercapX = cap_x[0] * pi / 180
        uppercapX = cap_x[1] * pi / 180
        cap_y = self.cap_y
        lowercapY = cap_y[0] * pi / 180
        uppercapY = cap_y[1] * pi / 180
        invert = self.invert
        smooth = 1 - (self.smoothing * .99)

        mouse_position = Vector(self.mouse.position)
        offset = (mouse_position - self.center) * -0.2

        if invert[1] is False:
            offset.y = -offset.y
        if invert[0] is True:
            offset.x = -offset.x
        offset *= sensitivity

        self._x = offset.x = interpolate(self._x, offset.x, smooth, 0)
        self._y = offset.y = interpolate(self._y, offset.y, smooth, 0)

        if use_cap_x:
            objectRotation = game_object_x.localOrientation.to_euler()

            if objectRotation.z + offset.x > uppercapX:
                offset.x = 0
                objectRotation.z = uppercapX
                game_object_x.localOrientation = objectRotation.to_matrix()

            if objectRotation.z + offset.x < lowercapX:
                offset.x = 0
                objectRotation.z = lowercapX
                game_object_x.localOrientation = objectRotation.to_matrix()

        game_object_x.applyRotation((0, 0, offset.x), self.local)

        rot_axis = 1 - self.front
        if use_cap_y:
            objectRotation = game_object_y.localOrientation.to_euler()

            if objectRotation[rot_axis] + offset.y > uppercapY:
                objectRotation[rot_axis] = uppercapY
                game_object_y.localOrientation = objectRotation.to_matrix()
                offset.y = 0

            if objectRotation[rot_axis] + offset.y < lowercapY:
                objectRotation[rot_axis] = lowercapY
                game_object_y.localOrientation = objectRotation.to_matrix()
                offset.y = 0

        rot = [0, 0, 0]
        rot[1-self.front] = offset.y
        game_object_y.applyRotation((*rot, ), True)
        if (Vector(self.mouse.position) - Vector(self.screen_center)).length > .00001:
        # if self.mouse.position != self.screen_center:
            self.mouse.position = self.screen_center
        self.done = True



from uplogic.nodes import ULActionNode
from uplogic.nodes import ULOutSocket
from uplogic.utils import is_waiting
from uplogic.utils import is_invalid
from uplogic.utils import not_met


class ULCopyProperty(ULActionNode):
    def __init__(self):
        ULActionNode.__init__(self)
        self.condition = None
        self.from_object = None
        self.to_object = None
        self.property_name = None
        self.done = False
        self.OUT = ULOutSocket(self, self._get_done)

    def _get_done(self):
        return self.done

    def evaluate(self):
        self.done = False
        condition = self.get_input(self.condition)
        if not_met(condition):
            return
        from_object = self.get_input(self.from_object)
        to_object = self.get_input(self.to_object)
        if is_invalid(from_object, to_object):
            return
        property_name = self.get_input(self.property_name)
        if is_waiting(property_name):
            return
        self._set_ready()
        val = from_object.get(property_name)
        if val is not None:
            to_object[property_name] = val
            self.done = True



'''TODO: Documentation
'''

from bge import logic
import time


def get_event_manager():
    scene = logic.getCurrentScene()
    if ULEventManager.update not in scene.post_draw:
        scene.post_draw.append(ULEventManager.update)
        # ULEventManager.initialized = True


class ULEventManager():
    '''Manager for `ULEvent` objects, not inteded for manual use.
    '''
    events = {}
    callbacks = []
    initialized = False
    done = []

    @classmethod
    def update(cls):
        for cb in cls.callbacks.copy():
            cb()

    @classmethod
    def log(cls):
        if cls.events:
            print('Events:')
            for evt in cls.events:
                print(f'\t{evt}:\t{cls.events[evt].content}')


    @classmethod
    def schedule(cls, cb):
        if not cls.initialized:
            get_event_manager()
        cls.callbacks.append(cb)

    @classmethod
    def deschedule(cls, cb):
        if not cls.initialized:
            get_event_manager()
        cls.callbacks.remove(cb)

    @classmethod
    def register(cls, event):
        if not cls.initialized:
            get_event_manager()
        cls.events[event.name] = event
        cls.schedule(event.remove)

    @classmethod
    def send(cls, name, content, messenger) -> None:
        if not cls.initialized:
            get_event_manager()
        ULEvent(name, content, messenger)


    @classmethod
    def receive(cls, name):
        if not cls.initialized:
            get_event_manager()
        return cls.events.get(name, None)

    @classmethod
    def consume(cls, name):
        if not cls.initialized:
            get_event_manager()
        return cls.events.pop(name, None)


class ULEvent():
    '''Event generated by `uplogic.events.send()`.
    
    **Not intended for manual use.**

    :param `name`: Name of the event; can be anything, not just `str`.
    :param `content`: This can be used to store data in an event.
    :param `messenger`: Can be used to store an object.
    '''

    def __init__(self, name, content=None, messenger=None):
        self.name = name
        self.content = content
        self.messenger = messenger
        ULEventManager.schedule(self.register)

    def register(self):
        ULEventManager.register(self)
        ULEventManager.deschedule(self.register)

    def remove(self):
        ULEventManager.events.pop(self.name, None)
        ULEventManager.deschedule(self.remove)


def send(name, content=None, messenger=None) -> None:
    '''Send an event that can be reacted to.

    :param `name`: Name of the event; can be anything, not just `str`.
    :param `content`: This can be used to store data in an event.
    :param `messenger`: Can be used to store an object.
    '''
    ULEventManager.send(name, content, messenger)


def receive(name) -> ULEvent:
    '''Check if an event has occured.

    :param `name`: Name of the event; can be anything, not just `str`.

    :returns: `ULEvent` with `name`, `content` and `messenger` as attributes.
    '''
    return ULEventManager.receive(name)


def consume(name: str):
    '''Check if an event has occured. This will remove the event.

    :param `name`: Name of the event; can be anything, not just `str`.

    :returns: `ULEvent` with `name`, `content` and `messenger` as attributes.
    '''
    return ULEventManager.consume(name, None)


def bind(name, callback):
    '''Send an event that can be reacted to.

    :param `name`: Name of the event; can be anything, not just `str`.
    :param `content`: This can be used to store data in an event.
    :param `messenger`: Can be used to store an object.
    '''
    def _check_evt(name, callback):
        evt = receive(name)
        if evt:
            callback(evt.name, evt.content, evt.messenger)
    logic.getCurrentScene().post_draw.append(_check_evt)


def schedule(name: str, content=None, messenger=None, delay=0.0):
    '''Send an event that can be reacted to with a delay.

    :param `name`: Name of the event; can be anything, not just `str`.
    :param `content`: This can be used to store data in an event.
    :param `messenger`: Can be used to store an object.
    :param `delay`: Delay with which to send the event in seconds.
    '''
    ScheduledEvent(delay, name, content, messenger)


class ScheduledEvent():
    '''Event generated by `uplogic.events.schedule()`.
    
    **Not intended for manual use.**

    :param `delay`: Delay with which to send the event in seconds.
    :param `name`: Name of the event; can be anything, not just `str`
    :param `content`: This can be used to store data in an event.
    :param `messenger`: Can be used to store an object.
    '''

    def __init__(self, delay, name, content, messenger):
        self.time = time.time()
        self.delay = self.time + delay
        self.name = name
        self.content = content
        self.messenger = messenger
        ULEventManager.schedule(self.send_scheduled)

    def send_scheduled(self):
        if time.time() >= self.delay:
            ULEventManager.deschedule(self.send_scheduled)
            ULEvent(self.name, self.content, self.messenger)


def schedule_callback(cb, delay=0.0, arg=None):
    '''Call a function with a delay. The function can have an argument when
    defined as a keyword.

    Callback cannot return anything.

    :param `cb`: Callback to be evaluated.
    :param `delay`: Delay with which to call the function in seconds.
    :param `arg`: If this is defined, callback will be called with this
    argument.
    '''
    ScheduledCallback(cb, delay, arg)


class ScheduledCallback():
    '''Event generated by `uplogic.events.schedule_callback()`.
    
    **Not intended for manual use.**

    :param `cb`: Callback to be evaluated.
    :param `delay`: Delay with which to call the function in seconds.
    :param `arg`: If this is defined, callback will be called with this
    argument.
    '''

    def __init__(self, cb, delay=0.0, arg=None):
        ULEventManager.schedule(self.call_scheduled)
        self.time = time.time()
        self.delay = self.time + delay
        self.callback = cb
        self.arg = arg

    def call_scheduled(self):
        if time.time() >= self.delay:
            if self.arg is not None:
                self.callback(self.arg)
            else:
                self.callback()
            ULEventManager.deschedule(self.call_scheduled)



from uplogic.nodes import ULParameterNode
from uplogic.nodes import ULOutSocket
from uplogic.utils import STATUS_WAITING, is_invalid
from mathutils import Vector


class ULScreenPosition(ULParameterNode):
    def __init__(self):
        ULParameterNode.__init__(self)
        self.game_object = None
        self.camera = None
        self.pos = None
        self.OUT = ULOutSocket(self, self.get_pos)

    def get_pos(self):
        game_object = self.get_input(self.game_object)
        camera = self.get_input(self.camera)
        if is_invalid(game_object) or is_invalid(camera):
            return STATUS_WAITING
        position = camera.getScreenPosition(game_object)
        self._set_value(position)
        return Vector((position[0], position[1]))

    def evaluate(self):
        self._set_ready()



from uplogic.nodes import ULParameterNode
from uplogic.nodes import ULOutSocket
from uplogic.utils import LOGIC_OPERATORS
from uplogic.utils import STATUS_WAITING
from uplogic.utils import is_waiting


class ULValueSwitch(ULParameterNode):
    def __init__(self):
        ULParameterNode.__init__(self)
        self.conditon = None
        self.val_a = None
        self.val_b = None
        self.out_value = False
        self.VAL = ULOutSocket(self, self._get_out_value)

    def _get_out_value(self):
        condition = self.get_input(self.condition)
        val_a = self.get_input(self.val_a)
        val_b = self.get_input(self.val_b)
        return (
            val_a if condition is True else val_b
        )

    def evaluate(self):
        self._set_ready()


class ULValueSwitchList(ULParameterNode):
    def __init__(self):
        ULParameterNode.__init__(self)
        self.ca = True
        self.val_a = None
        self.cb = True
        self.val_b = None
        self.cc = True
        self.val_c = None
        self.cd = True
        self.val_d = None
        self.ce = True
        self.val_e = None
        self.cf = True
        self.val_f = None
        self.out_value = False
        self.VAL = ULOutSocket(self, self._get_out_value)

    def _get_out_value(self):
        ca = self.get_input(self.ca)
        val_a = self.get_input(self.val_a)
        cb = self.get_input(self.cb)
        val_b = self.get_input(self.val_b)
        cc = self.get_input(self.cc)
        val_c = self.get_input(self.val_c)
        cd = self.get_input(self.cd)
        val_d = self.get_input(self.val_d)
        ce = self.get_input(self.ce)
        val_e = self.get_input(self.val_e)
        cf = self.get_input(self.cf)
        val_f = self.get_input(self.val_f)
        if is_waiting(ca, cb, cc, cd, ce, cf):
            return STATUS_WAITING
        if ca == True:
            return val_a
        elif cb == True:
            return val_b
        elif cc == True:
            return val_c
        elif cd == True:
            return val_d
        elif ce == True:
            return val_e
        elif cf == True:
            return val_f
        else:
            return

    def evaluate(self):
        self._set_ready()


class ULValueSwitchListCompare(ULParameterNode):
    def __init__(self, operator='GREATER'):
        ULParameterNode.__init__(self)
        self.operator = operator
        self.p0 = None
        self.val_default = None
        self.pa = None
        self.val_a = None
        self.pb = None
        self.val_b = None
        self.pc = None
        self.val_c = None
        self.pd = None
        self.val_d = None
        self.pe = None
        self.val_e = None
        self.pf = None
        self.val_f = None
        self.RESULT = ULOutSocket(self, self.get_result)

    def is_none_if_not_eq_or_neq_operator(self, a, b):
        if self.operator > 1:  # eq and neq are valid for None
            if a is None or b is None:
                return True
            else:
                return False
        return False

    def get_result(self):
        p0 = self.get_input(self.p0)
        val_default = self.get_input(self.val_default)
        pa = self.get_input(self.pa)
        val_a = self.get_input(self.val_a)
        pb = self.get_input(self.pb)
        val_b = self.get_input(self.val_b)
        pc = self.get_input(self.pc)
        val_c = self.get_input(self.val_c)
        pd = self.get_input(self.pd)
        val_d = self.get_input(self.val_d)
        pe = self.get_input(self.pe)
        val_e = self.get_input(self.val_e)
        pf = self.get_input(self.pf)
        val_f = self.get_input(self.val_f)
        operator = self.get_input(self.operator)
        if is_waiting(p0, pa, pb, pc, pd, pe, pf):
            return STATUS_WAITING
        if self.is_none_if_not_eq_or_neq_operator(p0, pa):
            return
        if LOGIC_OPERATORS[operator](p0, pa):
            return val_a
        if self.is_none_if_not_eq_or_neq_operator(p0, pb):
            return
        if LOGIC_OPERATORS[operator](p0, pb):
            return val_b
        if self.is_none_if_not_eq_or_neq_operator(p0, pc):
            return
        if LOGIC_OPERATORS[operator](p0, pc):
            return val_c
        if self.is_none_if_not_eq_or_neq_operator(p0, pd):
            return
        if LOGIC_OPERATORS[operator](p0, pd):
            return val_d
        if self.is_none_if_not_eq_or_neq_operator(p0, pe):
            return
        if LOGIC_OPERATORS[operator](p0, pe):
            return val_e
        if self.is_none_if_not_eq_or_neq_operator(p0, pf):
            return
        if LOGIC_OPERATORS[operator](p0, pf):
            return val_f
        else:
            return val_default

    def evaluate(self):
        self._set_ready()



from uplogic.audio import ULSound2D
from uplogic.nodes import ULActionNode
from uplogic.nodes import ULOutSocket
from uplogic.utils import is_invalid
from uplogic.utils import not_met


class ULStartSound(ULActionNode):
    def __init__(self):
        ULActionNode.__init__(self)
        self.condition = None
        self.sound = None
        self.loop_count = None
        self.pitch = None
        self.volume = None
        self.done = None
        self.device = None
        self.on_finish = False
        self._handle = None
        self.DONE = ULOutSocket(self, self.get_done)
        self.ON_FINISH = ULOutSocket(self, self.get_on_finish)
        self.HANDLE = ULOutSocket(self, self.get_handle)

    def get_handle(self):
        return self._handle

    def get_on_finish(self):
        if not self._handle:
            return False
        if self._handle.finished:
            return True
        return False

    def get_done(self):
        return self.done

    def reset(self):
        ULActionNode.reset(self)
        if self._handle and self._handle.finished:
            self._handle = None

    def evaluate(self):
        self.done = False
        self.on_finish = False
        self._set_ready()
        pitch = self.get_input(self.pitch)
        volume = self.get_input(self.volume)
        if self._handle:
            self._handle.pitch = pitch
            self._handle.volume = volume
        condition = self.get_input(self.condition)
        if not_met(condition):
            self._set_ready()
            return
        file = self.get_input(self.sound)
        loop_count = self.get_input(self.loop_count)

        if is_invalid(file):
            return

        self._handle = ULSound2D(
            file,
            'default',
            volume,
            pitch,
            loop_count
        )
        self.done = True



from uplogic.nodes import ULOutSocket
from uplogic.nodes import ULParameterNode
from uplogic.utils import STATUS_WAITING
from uplogic.utils import is_waiting


class ULInitNewDict(ULParameterNode):
    def __init__(self):
        ULParameterNode.__init__(self)
        self.key = None
        self.val = None
        self.DICT = ULOutSocket(self, self.get_dict)

    def get_dict(self):
        key = self.get_input(self.key)
        value = self.get_input(self.val)
        if is_waiting(key, value):
            return STATUS_WAITING
        return {str(key): value}

    def evaluate(self):
        self._set_ready()



from uplogic.nodes import ULConditionNode


class ULKeyboardActive(ULConditionNode):
    def __init__(self):
        ULConditionNode.__init__(self)

    def evaluate(self):
        self._set_ready()
        self._set_value(
            len(self.network.active_keyboard_events) > 0
        )



from uplogic.nodes import ULOutSocket
from uplogic.nodes import ULParameterNode
from uplogic.utils import STATUS_WAITING
from uplogic.utils import is_waiting
from uplogic.utils import compute_distance


class ULDistance(ULParameterNode):
    def __init__(self):
        ULParameterNode.__init__(self)
        self.parama = None
        self.paramb = None
        self.OUT = ULOutSocket(self, self.get_out)

    def get_out(self):
        socket = self.get_output('distance')
        if socket is None:
            parama = self.get_input(self.parama)
            paramb = self.get_input(self.paramb)
            if is_waiting(parama, paramb):
                return STATUS_WAITING
            return self.set_output(
                'distance',
                compute_distance(parama, paramb)
            )
        return socket

    def evaluate(self):
        self._set_ready()



from bge.types import KX_GameObject as GameObject
from bpy.types import NodeGroup
from bpy.types import Node
from bpy.types import NodeSocket
import bpy


def get_geom_socket(
    tree: str or NodeGroup,
    node: str or Node,
    socket: int or NodeSocket
):
    if isinstance(tree, str):
        tree = bpy.data.node_groups[tree]
    return (
        tree
        .nodes[node]
        .inputs[socket]
        .default_value
    )


def set_geom_socket(
    tree: str or NodeGroup,
    node: str or Node,
    socket: int or NodeSocket,
    value
):
    if isinstance(tree, str):
        tree = bpy.data.node_groups[tree]
    (
        tree
        .nodes[node]
        .inputs[socket]
        .default_value
    ) = value


def modify_geom_socket(
    tree: str or NodeGroup,
    node: str or Node,
    socket: int or NodeSocket,
    value
):
    if isinstance(tree, str):
        tree = bpy.data.node_groups[tree]
    (
        tree
        .nodes[node]
        .inputs[socket]
        .default_value
    ) += value


from bge import logic
import bpy
from bpy.types import Material


def create_curve(name, bevel_depth=0.0, dimensions=3, material=None, collection=None):
    bcurve = bpy.data.curves.new(name, 'CURVE')
    bcurve.bevel_depth = bevel_depth
    bcurve.dimensions = f'{dimensions}D'
    bobj = bpy.data.objects.new(name, bcurve)
    if material:
        if isinstance(material, str):
            bobj.data.materials.append(bpy.data.materials[material])
        elif isinstance(material, Material):
            bobj.data.materials.append(material)
    if collection:
        if isinstance(collection, str):
            collection = bpy.data.collections.get(collection, bpy.context.scene.collection)
    collection.objects.link(bobj)
    game_obj = logic.getCurrentScene().convertBlenderObject(bobj)
    return game_obj


def set_curve_points(curve, points):
    bcurve = curve.blenderObject.data
    for spline in bcurve.splines:
        bcurve.splines.remove(spline)
    spline = bcurve.splines.new('POLY')
    pos = curve.worldPosition

    spline.points.add(len(points)-1)
    for p, new_co in zip(spline.points, points):
        p.co = ([
            new_co[0] - pos.x,
            new_co[1] - pos.y,
            new_co[2] - pos.z
        ] + [1.0])



from uplogic.nodes import ULActionNode
from uplogic.nodes import ULOutSocket
from uplogic.utils import is_invalid
from uplogic.utils import is_waiting
from uplogic.utils import not_met


class ULAlignAxisToVector(ULActionNode):
    def __init__(self):
        ULActionNode.__init__(self)
        self.condition = None
        self.game_object = None
        self.vector = None
        self.axis = None
        self.factor = None
        self.OUT = ULOutSocket(self, self.get_done)

    def get_done(self):
        return self.done

    def evaluate(self):
        self.done = False
        condition = self.get_input(self.condition)
        self._set_ready()
        if not_met(condition):
            return
        game_object = self.get_input(self.game_object)
        v = self.get_input(self.vector)
        axis = self.get_input(self.axis)
        factor = self.get_input(self.factor)
        if is_invalid(game_object):
            return
        if is_waiting(axis, factor, v) or v.length <= 0:
            return
        v = getattr(v, 'worldPosition', v).copy()
        if not self.local:
            v -= game_object.worldPosition
        if axis > 2:
            matvec = v.copy()
            matvec.negate()
            v = matvec
            axis -= 3
        game_object.alignAxisToVect(v, axis, factor)
        self.done = True



from bge import logic
from bge.types import KX_GameObject as GameObject
from mathutils import Vector
from uplogic.nodes import ULActionNode
from uplogic.nodes import ULOutSocket
from uplogic.utils import is_invalid
from uplogic.utils import debug
from uplogic.utils import not_met


class ULGamepadLook(ULActionNode):
    def __init__(self, axis=0):
        ULActionNode.__init__(self)
        self.axis: int = axis
        self.condition = None
        self.main_obj: GameObject = None
        self.head_obj: GameObject = None
        self.inverted: bool = None
        self.index: int = None
        self.sensitivity: float = None
        self.exponent: float = None
        self.use_cap_x: bool = None
        self.cap_x: Vector = None
        self.use_cap_y: bool = None
        self.cap_y: Vector = None
        self.threshold: float = None
        self.done: bool = None
        self.DONE = ULOutSocket(self, self.get_done)

    def get_done(self):
        return self.done

    def evaluate(self):
        self.done = False
        axis: int = self.get_input(self.axis)
        condition: GameObject = self.get_input(self.condition)
        if not_met(condition):
            return
        main_obj: GameObject = self.get_input(self.main_obj)
        head_obj: GameObject = self.get_input(self.head_obj)
        if is_invalid(head_obj):
            head_obj = main_obj
        if is_invalid(axis):
            debug('Gamepad Sticks Node: Invalid Controller Stick!')
            return
        inverted: bool = self.get_input(self.inverted)
        index: int = self.get_input(self.index)
        sensitivity: float = self.get_input(self.sensitivity)
        exponent: float = self.get_input(self.exponent)
        threshold: float = self.get_input(self.threshold)
        use_cap_x: Vector = self.get_input(self.use_cap_x)
        cap_x: Vector = self.get_input(self.cap_x)
        uppercapX: float = cap_x.x
        lowercapX: float = -cap_x.y
        use_cap_y: Vector = self.get_input(self.use_cap_y)
        cap_y: Vector = self.get_input(self.cap_y)
        uppercapY: float = cap_y.x
        lowercapY: float = -cap_y.y

        self._set_ready()
        if logic.joysticks[index]:
            joystick = logic.joysticks[index]
        else:
            debug('Gamepad Sticks Node: No Joystick at that Index!')
            return
        if is_invalid(joystick):
            return
        raw_values = joystick.axisValues
        if axis == 0:
            x, y = raw_values[0], raw_values[1]
        elif axis == 1:
            x, y = raw_values[2], raw_values[3]
        neg_x = -1 if x < 0 else 1
        neg_y = -1 if y < 0 else 1

        if -threshold < x < threshold:
            x = 0
        else:
            x = abs(x) ** exponent

        if -threshold < y < threshold:
            y = 0
        else:
            y = abs(y) ** exponent
        if x == y == 0:
            self.done = True
            return

        x *= neg_x
        y *= neg_y

        x = -x if inverted['x'] else x
        y = -y if inverted['y'] else y

        x *= sensitivity
        if use_cap_x:
            objectRotation = main_obj.localOrientation.to_euler()
            if objectRotation.z + x > uppercapX:
                x = 0
                objectRotation.z = uppercapX
                main_obj.localOrientation = objectRotation.to_matrix()

            if objectRotation.z + x < lowercapX:
                x = 0
                objectRotation.z = lowercapX
                main_obj.localOrientation = objectRotation.to_matrix()

        y *= sensitivity
        if use_cap_y:
            objectRotation = head_obj.localOrientation.to_euler()
            if objectRotation.y + y > uppercapY:
                y = 0
                objectRotation.y = uppercapY
                head_obj.localOrientation = objectRotation.to_matrix()

            if objectRotation.y + y < lowercapY:
                y = 0
                objectRotation.y = lowercapY
                head_obj.localOrientation = objectRotation.to_matrix()

        main_obj.applyRotation(Vector((0, 0, x)), True)
        head_obj.applyRotation(Vector((0, y, 0)), True)
        self.done = True



from bge import logic
from uplogic.nodes import ULOutSocket
from uplogic.nodes import ULParameterNode


class ULGetGravity(ULParameterNode):
    def __init__(self):
        ULParameterNode.__init__(self)
        self.collection = None
        self.OUT = ULOutSocket(self, self.get_gravity)

    def get_gravity(self):
        return logic.getCurrentScene().gravity

    def evaluate(self):
        self._set_ready()



from uplogic.nodes import ULConditionNode
from uplogic.nodes import ULOutSocket
from uplogic.utils import STATUS_WAITING, is_invalid
from uplogic.utils import is_waiting


class ULLogicTreeStatus(ULConditionNode):
    def __init__(self):
        ULConditionNode.__init__(self)
        self.game_object = None
        self.tree_name = None
        self._running = False
        self._stopped = False
        self.tree = None
        self.IFRUNNING = ULOutSocket(self, self.get_running)
        self.IFSTOPPED = ULOutSocket(self, self.get_stopped)

    def get_running(self):
        socket = self.get_output('running')
        if socket is None:
            tree = self.tree
            if not tree:
                return STATUS_WAITING
            return self.set_output(
                'running',
                tree.is_running()
            )
        return socket

    def get_stopped(self):
        socket = self.get_output('stopped')
        if socket is None:
            tree = self.tree
            if not tree:
                return STATUS_WAITING
            return self.set_output(
                'stopped',
                tree.is_stopped()
            )
        return socket

    def evaluate(self):
        game_object = self.get_input(self.game_object)
        tree_name = self.get_input(self.tree_name)
        if is_waiting(game_object, tree_name):
            return
        self._set_ready()
        self._running = False
        self._stopped = False
        if is_invalid(game_object):
            return
        self.tree = game_object.get(f'IGNLTree_{tree_name}')



from .globaldb import GlobalDB
from .serializers import StringSerializer
from .serializers import FloatSerializer
from .serializers import IntegerSerializer
from .serializers import ListSerializer
from .serializers import VectorSerializer
from mathutils import Vector


GlobalDB.serializers[str(type(""))] = StringSerializer()
GlobalDB.serializers[str(type(1.0))] = FloatSerializer()
GlobalDB.serializers[str(type(10))] = IntegerSerializer()
GlobalDB.serializers[str(type([]))] = ListSerializer()
GlobalDB.serializers[str(type((0, 0, 0)))] = ListSerializer()
GlobalDB.serializers[str(type(Vector()))] = (
    VectorSerializer()
)



from uplogic.nodes import ULActionNode
from uplogic.nodes import ULOutSocket
from uplogic.utils import is_waiting
from uplogic.utils import not_met


class ULSetLightColor(ULActionNode):

    def __init__(self):
        ULActionNode.__init__(self)
        self.condition = None
        self.lamp = None
        self.color = None
        self.done = None
        self.OUT = ULOutSocket(self, self.get_done)

    def get_done(self):
        return self.done

    def evaluate(self):
        self.done = False
        condition = self.get_input(self.condition)
        if not_met(condition):
            self._set_value(False)
            return self._set_ready()
        lamp = self.get_input(self.lamp)
        color = self.get_input(self.color)
        if is_waiting(lamp, color):
            return
        if len(color) > 3:
            color = color[:-1]
        self._set_ready()
        light = lamp.blenderObject.data
        light.color = color
        self.done = True



from bge import logic


class ULCollision():
    """Collision Handler. Not intended for manual use."""
    target = None
    point = None
    normal = None
    tap = False
    consumed = False
    active = False
    _objects = []
    done_objs = []

    def __init__(self, game_object, callback, prop, mat, tap):
        self.callback = callback
        self.prop = prop
        self.mat = mat
        self.tap = tap
        self.game_object = game_object
        self.register()

    def collision(self, obj, point, normal):
        if self.tap and self.consumed:
            self.active = True
            return
        self._objects.append(obj)
        material = self.mat
        prop = self.prop
        bo = obj.blenderObject
        if material:
            if material not in [
                slot.material.name for
                slot in
                bo.material_slots
            ]:
                return
        if prop:
            for obj in self._objects:
                if prop not in obj:
                    return

        self.active = True
        if obj not in self.done_objs:
            self.callback(obj, point, normal)
            self.done_objs.append(obj)

    def reset(self):
        self.done_objs = []
        if not self.consumed and self.active:
            self.consumed = True
        elif self.consumed and not self.active:
            self.consumed = False
        self.active = False

    def register(self):
        if self.collision not in self.game_object.collisionCallbacks:
            self.game_object.collisionCallbacks.append(self.collision)
        logic.getCurrentScene().post_draw.append(self.reset)

    def unregister(self):
        self.game_object.collisionCallbacks.remove(self.collision)
        logic.getCurrentScene().post_draw.remove(self.reset)


def on_collision(obj, callback, prop='', material='', tap=False) -> ULCollision:
    """Bind a callback to an object's collision detection.

    :param `obj`: Object whose collision detection will be used.
    :param `callback`: Callback to be called when collision occurs. Must have arguments `(obj, point, norma)`.
    :param `prop`: Only look for objects that have this property.
    :param `material`: Only look for objects that have this material applied.
    :param `tap`: Only validate the first frame of the collision.
    """
    return ULCollision(obj, callback, prop, material, tap)



from .materials import get_material_socket
from .materials import set_material_socket
from .materials import get_world_socket
from .materials import set_world_socket
from .geometry import set_geom_socket
from .geometry import get_geom_socket
from .geometry import modify_geom_socket


from uplogic.nodes import ULOutSocket
from uplogic.nodes import ULParameterNode
from uplogic.utils import STATUS_WAITING
from uplogic.utils import is_invalid
import bpy


class ULGetCollectionObjectNames(ULParameterNode):
    def __init__(self):
        ULParameterNode.__init__(self)
        self.condition = None
        self.collection = None
        self.OUT = ULOutSocket(self, self.get_objects)

    def get_objects(self):
        socket = self.get_output('objects')
        if socket is None:
            collection = self.get_input(self.collection)
            if is_invalid(collection):
                return STATUS_WAITING
            col = bpy.data.collections.get(collection)
            if not col:
                return STATUS_WAITING
            objects = []
            for o in col.objects:
                if not o.parent:
                    objects.append(o.name)
            return self.set_output('objects', objects)
        return socket

    def evaluate(self):
        self._set_ready()



from uplogic.nodes import ULOutSocket
from uplogic.nodes import ULParameterNode
from uplogic.utils import LO_AXIS_TO_VECTOR
from uplogic.utils import STATUS_WAITING
from uplogic.utils import is_invalid


class ULAxisVector(ULParameterNode):
    def __init__(self):
        ULParameterNode.__init__(self)
        self.game_object = None
        self.OUT = ULOutSocket(self, self.get_vec)

    def get_vec(self):
        socket = self.get_output('vec')
        if socket is None:
            obj = self.get_input(self.game_object)
            front_vector = LO_AXIS_TO_VECTOR[self.axis]
            if is_invalid(obj, front_vector):
                return STATUS_WAITING
            return self.set_output(
                'vec',
                obj.getAxisVect(front_vector)
            )
        return socket

    def evaluate(self):
        self._set_ready()



from uplogic.animation import ULActionSystem
from uplogic.animation.action import ULAction
from uplogic.data import GlobalDB
from uplogic.nodes import ULActionNode
from uplogic.nodes import ULOutSocket
from uplogic.utils import is_invalid
from uplogic.utils import is_waiting
from uplogic.utils import not_met
import bpy


class ULSetActionFrame(ULActionNode):

    def __init__(self):
        ULActionNode.__init__(self)
        self.condition = None
        self.game_object = None
        self.action_layer = None
        self.action_frame = None
        self.action_name = None
        self.layer_weight = None
        self.freeze = None
        self.done = None
        self.act_system = self.get_act_system()
        self.OUT = ULOutSocket(self, self.get_done)
        
    def get_act_system(self):
        act_systems = GlobalDB.retrieve('uplogic.animation')
        if act_systems.check('default'):
            return act_systems.get('default')
        else:
            return ULActionSystem('default')

    def get_done(self):
        return self.done

    def evaluate(self):
        self.done = False
        condition = self.get_input(self.condition)
        if not_met(condition):
            self._set_ready()
            return
        game_object = self.get_input(self.game_object)
        action_layer = self.get_input(self.action_layer)
        action_frame = self.get_input(self.action_frame)
        freeze = self.get_input(self.freeze)
        action_name = self.get_input(self.action_name)
        layer_weight = self.get_input(self.layer_weight)
        self._set_ready()
        if is_waiting(
            action_layer,
            action_frame,
            layer_weight
        ):
            return
        if is_invalid(
            game_object,
        ):
            return

        action = self.act_system.get_layer(game_object, action_layer)
        same_action = action is not None and action.name == action_name
        if not same_action or action is None:
            action = bpy.data.actions[action_name]
            start_frame = action.frame_range[0]
            end_frame = action.frame_range[1]
            action = ULAction(
                game_object,
                action_name,
                start_frame,
                end_frame,
                action_layer,
                layer_weight=layer_weight
            )

        game_object.setActionFrame(action_frame, action_layer)
        if freeze:
            action.speed = 0
        action.layer_weight = layer_weight
        self.done = True



from bge import logic
from bge import render
from mathutils import Vector
from uplogic.nodes import ULActionNode
from uplogic.nodes import ULOutSocket
from uplogic.utils import is_waiting
from uplogic.utils import not_met


class ULProjectileRayCast(ULActionNode):

    def __init__(self):
        ULActionNode.__init__(self)
        self.condition = None
        self.origin = None
        self.destination = None
        self.power: float = None
        self.resolution: float = None
        self.property_name: str = None
        self.xray: bool = None
        self.distance: float = None
        self.visualize: bool = None
        self._picked_object = None
        self._point = None
        self._normal = None
        self._parabola = None
        self.PICKED_OBJECT = ULOutSocket(self, self.get_picked_object)
        self.POINT = ULOutSocket(self, self.get_point)
        self.NORMAL = ULOutSocket(self, self.get_normal)
        self.PARABOLA = ULOutSocket(self, self.get_parabola)
        self.network = None

    def setup(self, network):
        self.network = network

    def get_picked_object(self):
        return self._picked_object

    def get_parabola(self):
        return self._parabola

    def get_point(self):
        return self._point

    def get_normal(self):
        return self._normal

    def calc_projectile(self, t, vel, pos):
        half: float = logic.getCurrentScene().gravity.z * (.5 * t * t)
        vel = vel * t
        return Vector((0, 0, half)) + vel + pos

    def evaluate(self):
        condition = self.get_input(self.condition)
        if not_met(condition):
            self._set_value(False)
            self._out_normal = None
            self._out_object = None
            self._out_point = None
            return
        origin = self.get_input(self.origin)
        power: float = self.get_input(self.power)
        destination = self.get_input(self.destination)
        resolution: float = 1 - (self.get_input(self.resolution) * .99)
        property_name: str = self.get_input(self.property_name)
        xray: bool = self.get_input(self.xray)
        distance: float = self.get_input(self.distance)
        visualize: bool = self.get_input(self.visualize)

        if is_waiting(origin, destination, property_name, distance):
            return
        destination.normalize()
        destination *= power
        origin = getattr(origin, 'worldPosition', origin)

        points: list = []
        color: list = [1, 0, 0]
        idx = 0
        total_dist: float = 0
        found: bool = False
        owner = self.network._owner

        self._set_ready()

        while total_dist < distance:
            target = (self.calc_projectile(idx, destination, origin))
            start = origin if not points else points[-1]
            obj, point, normal = owner.rayCast(
                start,
                target,
                prop=property_name,
                xray=xray
            )
            total_dist += (target-start).length
            if not obj:
                points.append(target)
            else:
                points.append(point)
                color = [0, 1, 0]
                found = True
                break
            idx += resolution
        if visualize:
            for i, p in enumerate(points):
                if i < len(points) - 1:
                    render.drawLine(p, points[i+1], color)
        self._set_value(points[-1] if found else None)
        self._picked_object = obj
        self._point = point
        self._normal = normal
        self._parabola = points



from mathutils import Vector
from uplogic.nodes import ULConditionNode
from uplogic.utils import LOGIC_OPERATORS
from uplogic.utils import is_invalid
from uplogic.utils import is_waiting
from uplogic.utils import compute_distance


class ULCheckDistance(ULConditionNode):
    def __init__(self):
        ULConditionNode.__init__(self)
        self.param_a = None
        self.param_b = None
        self.operator = None
        self.dist = None
        self.hyst = None
        self._check = self._strict_check

    def _strict_check(self, opindex, value, hyst, threshold):
        v = LOGIC_OPERATORS[opindex](value, threshold)
        if v is True:
            self._set_value(True)
        else:
            self._set_value(False)
            self._check = self._hyst_check

    def _hyst_check(self, opindex, value, hyst, threshold):
        if (opindex == 2) or (opindex == 4):
            v = LOGIC_OPERATORS[opindex](value, threshold + hyst)
            if v is True:
                self._set_value(True)
                self._check = self._strict_check
            else:
                self._set_value(False)
        elif (opindex == 3) or (opindex == 5):
            v = LOGIC_OPERATORS[opindex](value, threshold - hyst)
            if v is True:
                self._set_value(True)
                self._check = self._strict_check
            else:
                self._set_value(False)

    def evaluate(self):
        a = self.get_input(self.param_a)
        b = self.get_input(self.param_b)
        op = self.get_input(self.operator)
        dist = self.get_input(self.dist)
        hyst = self.get_input(self.hyst)
        if is_waiting(a, b, op, dist, hyst):
            return
        self._set_ready()
        if is_invalid(a) or not isinstance(a, Vector):
            return
        if is_invalid(b) or not isinstance(b, Vector):
            return
        if op is None:
            return
        if dist is None:
            return
        distance = compute_distance(a, b)
        if distance is None:
            return self._set_value(False)
        if hyst is None:  # plain check, no threshold
            v = LOGIC_OPERATORS[op](distance, dist)
            self._set_value(v)
        else:  # check with hysteresys value, if >, >=, <, <=
            self._check(op, distance, hyst, dist)



from .actionsystem import ULActionSystem  # noqa
from .action import ACTION_FINISHED  # noqa
from .action import ACTION_STARTED  # noqa
from .action import ULAction  # noqa
from .sequence import ULSequence  # noqa



from uplogic.nodes import ULOutSocket
from uplogic.nodes import ULParameterNode
from uplogic.utils import is_invalid
from uplogic.utils import is_waiting


class ULListFromItems(ULParameterNode):
    def __init__(self):
        ULParameterNode.__init__(self)
        self.value = None
        self.value2 = None
        self.value3 = None
        self.value4 = None
        self.value5 = None
        self.value6 = None
        self.items: list = None
        self.LIST = ULOutSocket(self, self.get_list)

    def get_list(self):
        socket = self.get_output('list')
        if socket is None:
            value = self.get_input(self.value)
            value2 = self.get_input(self.value2)
            value3 = self.get_input(self.value3)
            value4 = self.get_input(self.value4)
            value5 = self.get_input(self.value5)
            value6 = self.get_input(self.value6)
            values = [value, value2, value3, value4, value5, value6]
            self.items = []
            for val in values:
                if not is_waiting(val) and not is_invalid(val):
                    self.items.append(val)
            return self.set_output('list', self.items)
        return socket

    def evaluate(self):
        self._set_ready()



from uplogic.nodes import ULActionNode
from uplogic.nodes import ULOutSocket
from uplogic.utils import is_invalid
from uplogic.utils import is_waiting
from uplogic.utils import not_met
import bpy


class ULSetMatNodeValue(ULActionNode):
    def __init__(self):
        ULActionNode.__init__(self)
        self.condition = None
        self.mat_name = None
        self.node_name = None
        self.internal = None
        self.attribute = None
        self.value = None
        self.done = False
        self.OUT = ULOutSocket(self, self._get_done)

    def _get_done(self):
        return self.done

    def evaluate(self):
        self.done = False
        condition = self.get_input(self.condition)
        if not_met(condition):
            self._set_ready()
            return
        mat_name = self.get_input(self.mat_name)
        node_name = self.get_input(self.node_name)
        attribute = self.get_input(self.attribute)
        internal = self.get_input(self.internal)
        value = self.get_input(self.value)
        if is_waiting(node_name, attribute, internal, value):
            return
        if is_invalid(mat_name):
            return
        if condition:
            self._set_ready()
            target = (
                bpy.data.materials[mat_name]
                .node_tree
                .nodes[node_name]
            )
            if internal:
                target = getattr(target, internal, target)
            if hasattr(target, attribute):
                setattr(target, attribute, value)
            self.done = True
